@@ -8,17 +8,23 @@
 
 from .data_model import KISAO_ALGORITHM_MAP
 from biosimulators_utils.combine.exec import exec_sedml_docs_in_archive
+from biosimulators_utils.config import get_config
 from biosimulators_utils.plot.data_model import PlotFormat  # noqa: F401
 from biosimulators_utils.report.data_model import ReportFormat, DataGeneratorVariableResults  # noqa: F401
 from biosimulators_utils.sedml.data_model import (Task, ModelLanguage, UniformTimeCourseSimulation,  # noqa: F401
                                                   DataGeneratorVariable, DataGeneratorVariableSymbol)
 from biosimulators_utils.utils.core import validate_str_value, parse_value
 from biosimulators_utils.sedml import validation
+from biosimulators_utils.simulator.data_model import AlgorithmSubstitutionPolicy
+from biosimulators_utils.simulator.exceptions import AlgorithmDoesNotSupportModelFeatureException
+from biosimulators_utils.simulator.warnings import AlgorithmSubstitutedWarning
+import numpy
 import os
 cwd = os.getcwd()  # because PySCeS changes the working directory
 import pysces  # noqa: E402
 os.chdir(cwd)
 import tempfile  # noqa: E402
+import warnings  # noqa: E402
 
 
 __all__ = ['exec_sedml_docs_in_combine_archive', 'exec_sed_task']
@@ -116,13 +122,24 @@ def exec_sed_task(task, variables):
         parsed_value = parse_value(value, setting['type'])
         model.__settings__[setting['id']] = parsed_value
 
+    # override algorithm choice if there are events
+    if integrator['id'] == 'LSODA' and model.__events__:
+        model.mode_integrator = 'CVODE'
+        if get_config().ALGORITHM_SUBSTITUTION_POLICY == AlgorithmSubstitutionPolicy.SAME_FRAMEWORK.value:
+            warnings.warn('CVODE (KISAO_0000019) will be used rather than LSODA (KISAO_0000088) because the model has events',
+                          AlgorithmSubstitutedWarning)
+        else:
+            raise AlgorithmDoesNotSupportModelFeatureException('LSODA cannot execute the simulation because the model has events')
+
     # setup time course
     model.sim_start = sim.initial_time
     model.sim_end = sim.output_end_time
-    model.sim_points = (sim.number_of_points
-                        * (sim.output_end_time - sim.initial_time)
-                        / (sim.output_end_time - sim.output_start_time)
-                        + 1)
+    model.sim_points = (
+        sim.number_of_points
+        * (sim.output_end_time - sim.initial_time)
+        / (sim.output_end_time - sim.output_start_time)
+        + 1
+    )
     if model.sim_points != int(model.sim_points):
         raise NotImplementedError('Time course must specify an integer number of time points')
 
@@ -135,11 +152,23 @@ def exec_sed_task(task, variables):
     unpredicted_targets = []
     results, labels = model.data_sim.getAllSimData(lbls=True)
     labels = {label: i_label for i_label, label in enumerate(labels)}
+
+    desired_time_points = numpy.linspace(sim.output_start_time, sim.output_end_time, sim.number_of_points + 1)
+    saved_time_points = model.data_sim.getTime()
+    time_step = (sim.output_end_time - sim.output_start_time) / sim.number_of_points
+    diff_saved_time_points = numpy.concatenate((numpy.diff(saved_time_points), numpy.array([time_step])))
+    i_times = []
+    eps = numpy.finfo(saved_time_points.dtype).eps * 2
+    for time in desired_time_points:
+        i_time = numpy.where(numpy.logical_and(saved_time_points >= time - eps, diff_saved_time_points > 0))[0][0]
+        i_times.append(i_time)
+    i_times = numpy.array(i_times)
+
     for variable in variables:
         if variable.symbol:
             if variable.symbol == DataGeneratorVariableSymbol.time:
                 i_result = labels['Time']
-                variable_results[variable.id] = results[:, i_result][-(sim.number_of_points + 1):]
+                variable_results[variable.id] = results[:, i_result][i_times]
             else:
                 unpredicted_symbols.append(variable.symbol)
 
@@ -149,7 +178,7 @@ def exec_sed_task(task, variables):
             if i_result is None:
                 unpredicted_targets.append(variable.target)
             else:
-                variable_results[variable.id] = results[:, i_result][-(sim.number_of_points + 1):]
+                variable_results[variable.id] = results[:, i_result][i_times]
 
     if unpredicted_symbols:
         raise NotImplementedError("".join([
